/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

/**
 * @brief Anything entity-related.
 * @file caosdb/entity.h
 * @author Timm Fitchen
 * @date 2021-07-07
 */
#ifndef CAOSDB_ENTITY_H
#define CAOSDB_ENTITY_H

#include "caosdb/message_code.h"
#include "caosdb/entity/v1alpha1/main.pb.h" // for Entity, RepeatedPtrField
#include "google/protobuf/util/json_util.h"
#include <memory> // for unique_ptr
#include <string> // for string

namespace caosdb::entity {

/**
 * Parent of an Entity.
 */
class Parent {
public:
  explicit inline Parent(caosdb::entity::v1alpha1::Parent *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] auto GetId() const -> const std::string &;
  [[nodiscard]] auto GetName() const -> const std::string &;
  [[nodiscard]] auto GetDescription() const -> const std::string &;

private:
  caosdb::entity::v1alpha1::Parent *wrapped;
};

/**
 * Container for parents of entities.
 */
class Parents {
public:
  inline Parents(){};
  explicit inline Parents(
    ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Parent>
      *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] auto At(int index) const -> const Parent &;

private:
  ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Parent>
    *wrapped;
  friend class Entity;
};

/**
 * Property of an Entity.
 *
 * This is a property which belongs to another entity. Don't confuse it with an
 * Entity with the "Property" role.
 *
 * @brief Property of an Entity.
 */
class Property {
public:
  explicit inline Property(caosdb::entity::v1alpha1::Property *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] auto GetId() const -> const std::string &;
  [[nodiscard]] auto GetName() const -> const std::string &;
  [[nodiscard]] auto GetDescription() const -> const std::string &;
  [[nodiscard]] auto GetImportance() const -> const std::string &;
  [[nodiscard]] auto GetValue() const -> const std::string &;
  [[nodiscard]] auto GetUnit() const -> const std::string &;
  [[nodiscard]] auto GetDatatype() const -> const std::string &;

private:
  caosdb::entity::v1alpha1::Property *wrapped;
};

/**
 * Container for Properties of Entities
 */
class Properties {
public:
  inline Properties(){};
  explicit inline Properties(
    ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Property>
      *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] auto At(int index) const -> const Property &;

private:
  ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Property>
    *wrapped;
  friend class Entity;
};

class Message {
public:
  explicit inline Message(caosdb::entity::v1alpha1::Message *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] inline auto GetCode() const -> MessageCode {
    return get_message_code(wrapped->code());
  }
  [[nodiscard]] inline auto GetDescription() const -> std::string {
    return wrapped->description();
  }

private:
  caosdb::entity::v1alpha1::Message *wrapped;
};

/**
 * Container for Messages.
 */
class Messages {
public:
  inline Messages() : wrapped(nullptr){};
  explicit inline Messages(
    ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Message>
      *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] inline auto Size() const -> int { return wrapped->size(); }
  [[nodiscard]] inline auto At(int index) const -> const Message {
    return Message(&(wrapped->at(index)));
  }

private:
  ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Message>
    *wrapped;
  friend class Entity;
};

/**
 * @brief Wrapper for the Protobuf entity.
 */
class Entity {
public:
  explicit inline Entity(caosdb::entity::v1alpha1::Entity *wrapped)
    : wrapped(wrapped) {
    errors.wrapped = this->wrapped->mutable_errors();
  };

  [[nodiscard]] inline auto GetId() const -> const std::string & {
    return wrapped->id();
  };
  [[nodiscard]] inline auto GetVersion() const -> const std::string & {
    return wrapped->version().id();
  };

  [[nodiscard]] inline auto GetRole() const -> const std::string & {
    return wrapped->role();
  };
  [[nodiscard]] inline auto GetName() const -> const std::string & {
    return wrapped->name();
  };
  [[nodiscard]] inline auto GetDescription() const -> const std::string & {
    return wrapped->description();
  };

  [[nodiscard]] auto GetDatatype() const -> const std::string & {
    return wrapped->datatype();
  };
  [[nodiscard]] auto GetUnit() const -> const std::string & {
    return wrapped->unit();
  };

  [[nodiscard]] auto GetParents() const -> const Parents &;
  [[nodiscard]] auto GetProperties() const -> const Properties &;
  [[nodiscard]] inline auto GetErrors() const -> const Messages & {
    return errors;
  }
  [[nodiscard]] inline auto HasErrors() const -> bool {
    return this->errors.wrapped->size() > 0;
  }

  inline auto ToString() const -> const std::string {
    google::protobuf::util::JsonOptions options;
    std::string out;
    google::protobuf::util::MessageToJsonString(*(this->wrapped.get()), &out,
                                                options);
    return out;
  }

private:
  std::unique_ptr<caosdb::entity::v1alpha1::Entity> wrapped;
  Properties properties;
  Parents parents;
  Messages errors;
};

} // namespace caosdb::entity
#endif
